from io import BytesIO
from contextlib import closing
from .mail import Mail
from email.parser import Parser


class POP3Parser:
    """
    RFC: https://datatracker.ietf.org/doc/html/rfc1939

    USER [username] 处理用户名
    PASS [password] 处理用户密码
    APOP [Name,Digest] 认可Digest是MD5消息摘要
    STAT 处理请求服务器发回关于邮箱的统计资料，如邮件总数和总字节数
    UIDL [邮件id] 处理返回邮件的唯一标识符，POP3会话的每个标识符都将是唯一的
    LIST [邮件id] 处理返回邮件数量和每个邮件的大小
    RETR [邮件id] 处理返回由参数标识的邮件的全部文本
    DELE [邮件id] 处理服务器将由参数标识的邮件标记为删除，最后由【quit】命令执行
    RSET 处理服务器将重置所有标记为删除的邮件，用于撤消DELE命令
    TOP [邮件id n] 处理服务器将返回由参数标识的邮件前n行内容，n必须是正整数
    NOOP 处理服务器返回一个肯定的响应
    QUIT 终止会话
    """

    def _parseEmail(self, content: bytes) -> Mail:
        pass

    def parse(self, responseBytes: bytes) -> [Mail]:
        """
        解析 TCP 流，从其中解析出邮件
        :param responseBytes:   传入POP3响应，源端口为110
        :return:
        """
        emails = []
        response = []
        # +OK
        lines = []
        with closing(BytesIO(responseBytes)) as data:
            lines = data.readlines()
        start = -1
        for idx, line in enumerate(lines):
            if line.startswith(b'+OK') or line.startswith(b'-ERR'):
                if idx >= 0:
                    tmp = b''.join(lines[start:idx])
                    tmpIdx = tmp.find(b'\r\n')
                    if tmpIdx != -1:
                        tmp = tmp[tmpIdx + 2:]
                    mail = Mail()
                    msg = Parser().parsestr(tmp.decode('utf-8'))
                    if msg.get("From") is not None and msg.get("To") is not None:
                        mail.parse(msg)
                        emails.append(mail)
                    response.append(tmp)
                start = idx
        return emails



from dpkt import ip, ethernet, tcp, udp, http, UnpackError, ip6
from datetime import datetime
from .FlowBase import FlowBase
from io import BytesIO
from contextlib import closing


class TCPFlow(FlowBase):
    """
    表示一个 TCP 流
    """

    def __init__(self, ethPacket: ethernet.Ethernet, timestamp: float):
        self.http_requests = []
        self.forwardPackets = dict()  # 一个字典，用来根据序列号记录正向TCP网络包
        self.reversePackets = dict()  # 一个字典，用来根据序列号记录反向TCP网络包
        self.forwardClosed = False  # 正向流是否关闭
        self.reverseClosed = False  # 反向流是否关闭
        super().__init__(ethPacket, timestamp)

    def addForwardPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        """
        添加一个正向网络包
        :param ethPacket:
        :param timestamp:
        :return:
        """
        # 如果不是TCP或者UDP包，则忽略
        if not FlowBase.canBeMarkAsFlow(ethPacket):
            return

        super(TCPFlow, self)._addForwardPacket(ethPacket, timestamp)

        tcpPacket = ethPacket.data.data
        if tcpPacket.dport in ["993", "995", "443"]:
            return

        if len(tcpPacket.data) == 0:
            return

        # 将TCP包存到字典里面，后面再对网络包进行排序可以还原TCP流的内容
        if tcpPacket.seq not in self.forwardPackets:
            self.forwardPackets[tcpPacket.seq] = tcpPacket

        # # 尝试从中解析出 http 请求信息
        # try:
        #     http_req = http.Request(tcpPacket.data)
        #     self.http_requests.append(http_req)
        # except UnpackError:
        #     pass

    def addReversePacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        """
        添加一个反向网络包
        :param ethPacket:
        :param timestamp:
        :return:
        """
        # 如果不是TCP或者UDP包，则忽略
        if not FlowBase.canBeMarkAsFlow(ethPacket):
            return
        super(TCPFlow, self)._addReversePacket(ethPacket, timestamp)

        tcpPacket = ethPacket.data.data

        if tcpPacket.sport in ["993", "995", "443"]:
            return

        if len(tcpPacket.data) == 0:
            return

        # 将TCP包存到字典里面，后面再对网络包进行排序可以还原TCP流的内容
        if tcpPacket.seq not in self.reversePackets:
            self.reversePackets[tcpPacket.seq] = tcpPacket

    def getAllForwardBytes(self) -> bytes:
        """
        返回所有正向流的集合
        :return:
        """
        with closing(BytesIO()) as byteBuffer:
            for packet in self.forwardIter():
                byteBuffer.write(packet.data)
            return byteBuffer.getvalue()

    def getAllReverseBytes(self) -> bytes:
        """
        返回所有反向流的集合
        :return:
        """
        with closing(BytesIO()) as byteBuffer:
            for packet in self.reverseIter():
                byteBuffer.write(packet.data)
            return byteBuffer.getvalue()

    def forwardIter(self):
        """
        自定义正向流迭代器，按序列号顺序遍历当前 TCP 流的所有的 TCP 包
        [1, 3, 7, 10]
        :return:
        """
        sortedPackets = sorted(list(self.forwardPackets.values()), key=lambda v: v.seq)
        for packet in sortedPackets:
            yield packet

    def reverseIter(self):
        """
        自定义反向流迭代器，按序列号顺序遍历当前 TCP 流的所有的 TCP 包
        :return:
        """
        sortedPackets = sorted(list(self.reversePackets.values()), key=lambda v: v.seq)
        for packet in sortedPackets:
            yield packet

    @staticmethod
    def _isValid(myIter):
        nextSeq = None
        for packet in myIter:
            if nextSeq is None:
                nextSeq = packet.seq + len(packet.data)
                continue
            if nextSeq != packet.seq:
                return False
            if len(packet.data) == 0:
                nextSeq += 1
            else:
                nextSeq += len(packet.data)
        return True

    def isForwardValid(self):
        """
        判断当前正向 TCP 流是否是有效的流
        :return:
        """
        if len(self.forwardPackets) == 0:
            return True
        if self.forwardClosed:
            return True
        return TCPFlow._isValid(self.forwardIter())

    def isReverseValid(self):
        """
        判断当前反向 TCP 流是否是有效的流
        :return:
        """
        if len(self.reversePackets) == 0:
            return True
        if self.reverseClosed:
            return True
        return TCPFlow._isValid(self.reverseIter())

    def isValid(self):
        """
        判断当前 TCP 流是否完整
        :return:
        """
        if not (self.forwardClosed and self.reverseClosed):
            return False
        return self.isForwardValid() and self.isReverseValid()

    def toFlowRecord(self):
        if self.protocol == ip.IP_PROTO_TCP:
            protocolStr = "tcp"
        else:
            protocolStr = "udp"
        return {
            "src_ip": self.srcIP,
            "dst_ip": self.dstIP,
            "src_port": self.srcPort,
            "dst_port": self.dstPort,
            "type": self.category,
            "protocol": protocolStr,
            "start_time": datetime.fromtimestamp(self.startTime),
            "end_time": datetime.fromtimestamp(self.lastTime),
        }



import gc

from dpkt import ethernet, ip, tcp

from .TCPFlow import TCPFlow
from .FlowBase import FlowBase
from .Extractor import Extractor


class FlowExtractor(Extractor):
    # 6-192.168.1.4:13853-192.168.1.5:80 => Flow1
    # 6-192.168.1.5:80-192.168.1.4:13853 => Flow1
    flowMap = {}

    def __init__(self, valueCallback):
        super().__init__(valueCallback)
        self.lastCleanTime = 0
        self.flowMap = {}

    def addPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        # 只处理 TCP 和 UDP 包，其它包直接忽略（同时支持处理 IPv4 和 IPv6）
        if not FlowBase.canBeMarkAsFlow(ethPacket):
            return
        # 用四元组（源IP，源端口，目的IP，目的端口标识一条流）
        key = FlowBase.encodeToStr(ethPacket)
        reverseKey = FlowBase.encodeToReverseStr(ethPacket)

        ipPacket = ethPacket.data
        protocol = FlowBase.getProtocol(ethPacket)
        if protocol == ip.IP_PROTO_TCP:
            tcpPacket = ipPacket.data

            # 判断是否已经前面有该流量的记录
            if key in self.flowMap:
                tcpFlow = self.flowMap[key]
                forward = key == tcpFlow.forwardStr
                if forward:
                    tcpFlow.addForwardPacket(ethPacket, timestamp)
                    # check if flow needs to be closed due to fin flag and verify stream
                    if tcpPacket.flags & tcp.TH_FIN:
                        if tcpFlow.isForwardValid():
                            tcpFlow.forwardClosed = True
                        del self.flowMap[key]
                else:
                    tcpFlow.addReversePacket(ethPacket, timestamp)
                    # check if flow needs to be closed due to fin flag and verify stream
                    if tcpPacket.flags & tcp.TH_FIN:
                        if tcpFlow.isReverseValid():
                            tcpFlow.reverseClosed = True
                        del self.flowMap[key]
                if tcpFlow.isValid():
                    self.valueCallback(tcpFlow)

            else:
                # TCP 流不存在的情况下，收到一个 SYN 包创建一个新的 TCP 流
                if not tcpPacket.flags & tcp.TH_SYN:
                    return
                self.flowMap[key] = TCPFlow(ethPacket, timestamp)
                self.flowMap[reverseKey] = self.flowMap[key]

        elif protocol == ip.IP_PROTO_UDP:
            # 处理 UDP 包
            udpPacket = ipPacket.data
            pass

        # 30s清理一次
        if timestamp - self.lastCleanTime > 30:
            self.lastCleanTime = timestamp
            # 清理所有5分钟内没有收到新的包的连接
            needCleanKeys = []
            for key in self.flowMap.keys():
                if timestamp - self.flowMap[key].lastTime > 15:
                    needCleanKeys.append(key)
                elif timestamp - self.flowMap[key].startTime > 15 and self.flowMap[key].totalPayloadBytes == 0:
                    needCleanKeys.append(key)

            for cleanKey in needCleanKeys:
                del self.flowMap[cleanKey]
            gc.collect()

    def done(self):
        pass
        # for key in self.flowMap:
        #     flow = self.flowMap[key]
        #     self.valueCallback(flow)



import hashlib
import json
from dpkt import http, UnpackError
from .MyEncoder import MyEncoder


class HttpData:
    def __init__(self):
        self.request = None
        self.response = None
        self.fileHashes = None

    def getDomain(self):
        if self.request.headers:
            if "host" in self.request.headers and self.request.headers["host"]:
                return self.request.headers["host"].strip()
        return ""

    def getUrl(self):
        return f"http://{self.getDomain()}{self.request.uri}"

    def __repr__(self):
        request = self.request
        response = self.response
        if request is None:
            return "None"
        requestDict = {

        }
        if request is not None:
            requestDict = {
                "uri": request.uri,
                "method": request.method,
                "headers": request.headers,
                "version": request.version,
                "body": request.body,
            }
        responseDict = {

        }
        if response is not None:
            responseDict = {
                "status": response.status,
                "reason": response.reason,
                "body": response.body,
                "headers": response.headers,
                "version": response.version,
            }
        return json.dumps({
            "request": requestDict,
            "response": responseDict,
            "fileHash": self.fileHashes
        }, ensure_ascii=False, cls=MyEncoder)


class HTTPParser:
    def parse(self, requestBytes, responseBytes: bytes) -> [HttpData]:
        # extract request
        startIdx = 0
        requests, responses = [], []
        while startIdx < len(requestBytes):
            try:
                http_req = http.Request(requestBytes[startIdx:])
                # http_req.headers = dict(http_req.headers)
                startIdx += len(http_req.__bytes__())
                requests.append(http_req)
            except UnpackError as e:
                # print(e)
                return []
        # extract response
        if len(requests) == 0:
            return []
        startIdx = 0
        while startIdx < len(responseBytes):
            try:
                http_response = http.Response(responseBytes[startIdx:])
                startIdx += len(http_response.__bytes__())
                responses.append(http_response)
            except UnpackError as e:
                # print(e)
                return []
        if len(requests) != len(responses):
            print(f"Request num: {len(requests)}, Response num: {len(responses)}, Not match")
            return []
        httpDataArr = []
        for i, http_req in enumerate(requests):
            httpData = HttpData()
            httpData.request = http_req
            httpData.response = responses[i]
            if 'content-type' in httpData.response.headers:
                contentType = httpData.response.headers['content-type'].strip()
                if contentType.startswith("image") or contentType.startswith("audio") or contentType.startswith(
                        "video") or \
                        contentType.startswith("application"):
                    httpData.fileHashes = {
                        "md5": hashlib.md5(httpData.response.body).hexdigest(),
                        "sha1": hashlib.sha1(httpData.response.body).hexdigest(),
                        "sha256": hashlib.sha256(httpData.response.body).hexdigest()
                    }
                    httpData.response.body = "It's a file, already cal file hashes"
            httpDataArr.append(httpData)
        return httpDataArr



import csv
import datetime
import hashlib
import os

from dpkt import ethernet, ip, ip6
from dpkt.utils import inet_to_str
from .pcap_extractor.mail import Mail
from .pcap_extractor.TCPFlow import TCPFlow
from .pcap_extractor.DNSExtractor import DNSItem
from .pcap_extractor.HTTPParser import HttpData
from abc import ABCMeta, abstractmethod


class DictSerializable(object):
    """
    可序列化为字典的类的基类 => 继承本抽象类的类拥有序列化成字典的能力
    """
    __metaclass__ = ABCMeta  # 指定这是一个抽象类

    @abstractmethod
    def toDict(self) -> dict:
        pass


class TCPMixin(DictSerializable):
    """
    TCP记录，所有继承本类的记录都拥有记录TCP属性的功能
    """

    def __init__(self):
        self.srcIP = ""
        self.srcPort = 0
        self.dstIP = ""
        self.dstPort = 0
        self.startTime = 0
        self.endTime = 0

    def setTCP(self, tcpFlow: TCPFlow):
        self.srcIP = tcpFlow.srcIP
        self.srcPort = tcpFlow.srcPort
        self.dstIP = tcpFlow.dstIP
        self.dstPort = tcpFlow.dstPort
        self.startTime = tcpFlow.startTime
        self.endTime = tcpFlow.lastTime

    def toDict(self) -> dict:
        return {
            "srcIP": self.srcIP,
            "srcPort": self.srcPort,
            "dstIP": self.dstIP,
            "dstPort": self.dstPort
        }


class IPMixin(DictSerializable):
    """
    IP记录，所有继承本类的记录都拥有记录IP地址的功能
    """

    def __init__(self):
        self.srcIP = ""
        self.dstIP = ""

    def setIP(self, ethPacket: ethernet.Ethernet):
        ipPacket = ethPacket.data
        self.srcIP = inet_to_str(ipPacket.src)
        self.dstIP = inet_to_str(ipPacket.dst)

    def toDict(self) -> dict:
        return {
            "srcIP": self.srcIP,
            "dstIP": self.dstIP
        }


class RecordBase(DictSerializable):
    """
    所有报告记录的基类
    """

    __metaclass__ = ABCMeta  # 指定这是一个抽象类

    def __init__(self):
        self.observables = []

    def setObservables(self, observables: [str]):
        self.observables = observables

    def toDict(self) -> dict:
        return {
            "observables": self.observables
        }


class FileHash(DictSerializable):
    """
    表示文件Hash值
    """

    def __init__(self, data: bytes = None):
        if data:
            self.md5 = hashlib.md5(data).hexdigest()
            self.sha1 = hashlib.sha1(data).hexdigest()
            self.sha256 = hashlib.sha256(data).hexdigest()
        else:
            self.md5 = ""
            self.sha1 = ""
            self.sha256 = ""

    def toDict(self) -> dict:
        return {
            "md5": self.md5,
            "sha1": self.sha1,
            "sha256": self.sha256
        }


class FileItem(DictSerializable):
    """
    文件条目
    """

    def __init__(self, fileData: bytes, filename: str = "", fileType: str = ""):
        self.filename = filename
        self.fileType = fileType
        self.fileHash = FileHash(fileData)

    def toDict(self) -> dict:
        return {
            "filename": self.filename,
            "fileType": self.fileType,
            "fileHash": self.fileHash.toDict()
        }


class EmailRecord(RecordBase, TCPMixin):
    """
    Email记录
    """

    def __init__(self):
        super().__init__()
        self.plain = ""  # 消息内容
        self.html = ""  # html格式的内容
        self.From = ""  # 发件人地址
        self.To = ""  # 收件人地址
        self.Cc = ""  # 抄送地址
        # self.Date = ""  # 日期和时间
        self.Subject = ""  # 主题
        self.MessageID = ""  # 消息ID
        self.files = []

    def initByMailAndObservable(self, mail: Mail):
        self.plain = mail.plain
        if isinstance(self.plain, bytes):
            self.plain = self.plain.decode()
        self.html = mail.html
        self.From = mail.From
        self.To = mail.To
        self.Cc = mail.Cc
        # self.Date = mail.Date
        self.Subject = mail.Subject
        self.MessageID = mail.MessageID
        for mailFile in mail.files:
            self.files.append(FileItem(mailFile.fileData, mailFile.fileName, mailFile.fileType))

    def toDict(self) -> dict:
        filesDict = []
        for file in self.files:
            filesDict.append(file.toDict())
        return {
            "plain": self.plain,
            "html": self.html,
            "from": self.From,
            "to": self.To,
            "cc": self.Cc,
            # "date": self.Date,
            "subject": self.Subject,
            "messageId": self.MessageID,
            "files": filesDict,
            **RecordBase.toDict(self),
            **TCPMixin.toDict(self)
        }


class DNSRecord(RecordBase, IPMixin):
    """
    DNS解析记录
    """

    def __init__(self):
        super().__init__()
        self.domain = ""
        self.domain_type = ""
        self.value = ""
        self.timestamp = 0

    def initByDNSItem(self, item: DNSItem):
        self.domain = item.domain
        self.domain_type = item.domain_type
        self.value = item.value
        self.timestamp = item.timestamp
        self.setIP(item.ethPacket)

    def toDict(self) -> dict:
        return {
            "domain": self.domain,
            "domain_type": self.domain_type,
            "value": self.value,
            "timestamp": self.timestamp,
            **RecordBase.toDict(self),
            **IPMixin.toDict(self)
        }


class HTTPRecord(RecordBase, TCPMixin):
    """
    HTTP记录
    """

    class HttpRequest(DictSerializable):
        """
        HTTP请求记录
        """

        def __init__(self, httpData: HttpData):
            self.uri = httpData.request.uri
            self.method = httpData.request.method
            self.headers = httpData.request.headers
            self.version = httpData.request.version
            self.domain = httpData.getDomain()
            self.url = httpData.getUrl()

        def toDict(self) -> dict:
            return {
                "uri": self.uri,
                "method": self.method,
                "headers": self.headers,
                "version": self.version,
                "domain": self.domain,
                "url": self.url
            }

    class HttpResponse(DictSerializable):
        """
        HTTP响应记录
        """

        def __init__(self, httpData: HttpData):
            self.status = httpData.response.status
            self.reason = httpData.response.reason
            self.body = httpData.response.body
            self.headers = httpData.response.headers
            self.version = httpData.response.version

        def toDict(self) -> dict:
            return {
                "status": self.status,
                "reason": self.reason,
                # "body": self.body,
                "headers": self.headers,
                "version": self.version
            }

    def __init__(self, httpData: HttpData):
        super().__init__()
        self.request = HTTPRecord.HttpRequest(httpData)
        self.response = HTTPRecord.HttpResponse(httpData)

    def toDict(self) -> dict:
        return {
            "request": self.request.toDict(),
            "response": self.response.toDict(),
            **RecordBase.toDict(self),
            **TCPMixin.toDict(self)
        }


class FileHashRecord(RecordBase, TCPMixin):
    """
    FTP 记录
    """

    def __init__(self, data: bytes = None):
        super().__init__()
        self.fileHash = None
        self.filename = ""
        if data:
            self.fileHash = FileHash(data)

    def getValue(self):
        return f"{self.fileHash.md5}-{self.fileHash.sha1}-{self.fileHash.sha256}"

    def toDict(self) -> dict:
        return {
            "fileHash": self.fileHash.toDict(),
            **RecordBase.toDict(self),
            **TCPMixin.toDict(self)
        }


class Report(DictSerializable):
    """
    对比报告
    """

    def toDict(self) -> dict:
        emailRecordsDict = []
        fileHashRecordsDict = []
        httpRecordsDict = []
        domainRecordsDict = []
        for emailRecord in self.emailRecords:
            emailRecordsDict.append(emailRecord.toDict())
        for fileHashRecord in self.fileHashRecords:
            fileHashRecordsDict.append(fileHashRecord.toDict())
        for httpRecord in self.httpRecords:
            httpRecordsDict.append(httpRecord.toDict())
        for domainRecord in self.domainRecords:
            domainRecordsDict.append(domainRecord.toDict())
        return {
            "totalPacket": self.totalPacket,
            "totalIPAddress": self.totalIPAddress,
            "totalIPv6Address": self.totalIPv6Address,
            "totalIPPacket": self.totalIPPacket,
            "totalIPv6Packet": self.totalIPv6Packet,
            "duration": self.duration,
            "totalTCPFlowNum": self.totalTCPFlowNum,
            "totalHTTPNum": self.totalHTTPNum,
            "totalFTPNum": self.totalFTPNum,
            "totalEmailNum": self.totalEmailNum,
            "totalFileNum": self.totalFileNum,
            "totalDomainNum": self.totalDomainNum,
            "totalMatchIpAddress": self.totalMatchIpAddress,
            "totalMatchIpv6Address": self.totalMatchIpv6Address,
            "totalMatchEmailNum": self.totalMatchEmailNum,
            "totalMatchDomain": self.totalMatchDomain,
            "totalMatchFileHash": self.totalMatchFileHash,
            "totalMatchUri": self.totalMatchUri,
            "startTime": self.startTime,
            "endTime": self.endTime,
            "observables": self.observablesMap,
            "emailRecords": emailRecordsDict,
            "fileHashRecords": fileHashRecordsDict,
            "httpRecords": httpRecordsDict,
            "domainRecords": domainRecordsDict,
        }

    def __init__(self):
        self.totalPacket = 0  # Total packet num in pcap file
        self.totalIPAddress = 0  # Total ipv4 address num in pcap file
        self.totalIPv6Address = 0  # Total ipv6 address num in pcap file
        self.totalEmailAddress = 0  # Total email address num in pcap file
        self.totalIPPacket = 0  # Total ipv4 packet num in pcap file
        self.totalIPv6Packet = 0  # Total ipv6 pcaket num in pcap file
        self.duration = 0  # The duration between the begin of pcap file and end of pcap file
        self.totalTCPFlowNum = 0  # Total TCP flow num in pcap file
        self.totalHTTPNum = 0  # Total HTTP session num in pcap file
        self.totalFTPNum = 0  # Total FTP session num in pcap file
        self.totalEmailNum = 0  # Total email num in pcap file，contain SMTP、POP3 and IMAP protocol
        self.totalFileNum = 0  # Total file num extract from pcap file, from FTP、SMTP、POP3 、IMAP and HTTP
        self.totalDomainNum = 0  # Total domain num extract from DNS query and DNS response）

        self.totalMatchIpAddress = 0  # Total ipv4 address matched
        self.totalMatchIpv6Address = 0  # Total ipv6 address matched
        self.totalMatchEmailNum = 0  # Total email matched
        self.totalMatchDomain = 0  # Total domain matched
        self.totalMatchFileHash = 0  # Total file hash matched
        self.totalMatchUri = 0  # Total match uri

        self.emailRecords = []  # contain all email which some observable match it
        self.fileHashRecords = []  # contain all FTP session which some observable match it
        self.httpRecords = []  # contain all HTTP session which some observable match it
        self.domainRecords = []  # contain all Domain query record which some observable match it
        self.observablesMap = dict()  # 一个字典，用来保存 Observable => <observableId, observable>

        self.isFirst = True
        self.isDumpCsv = True
        self.ipv4AddressSet = set()
        self.ipv6AddressSet = set()
        self.emailAddressSet = set()
        self.startTime = 0
        self.endTime = 0
        self.outputCsvDir = "outputcsv"
        self.pcapFile = ""

        self.ipAddressCSVFile = None
        self.emailAddressCSVFile = None
        self.uriCSVFile = None
        self.fileHashCSVFile = None
        self.domainCSVFile = None

        self.ipAddressCSVWriter = None
        self.emailAddressCSVWriter = None
        self.uriCSVWriter = None
        self.fileHashCSVWriter = None
        self.domainCSVWriter = None

    def begin(self):
        """
        报告开始
        :return:
        """
        pass

    def addPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        """
        每解析到一个以太网包，
        :return:
        """
        if self.isFirst:
            self.startTime = timestamp
            self.isFirst = False
        self.endTime = timestamp
        self.duration = self.endTime - self.startTime

        self.totalPacket += 1
        ipPacket = ethPacket.data
        if isinstance(ethPacket.data, ip.IP):
            # 如果是 IPv4 包
            self.totalIPPacket += 1
            self.ipv4AddressSet.add(inet_to_str(ipPacket.src))
            self.ipv4AddressSet.add(inet_to_str(ipPacket.dst))
            self.totalIPAddress = len(self.ipv4AddressSet)
        elif isinstance(ethPacket.data, ip6.IP6):
            self.totalIPv6Packet += 1
            self.ipv6AddressSet.add(inet_to_str(ipPacket.src))
            self.ipv6AddressSet.add(inet_to_str(ipPacket.dst))
            self.totalIPv6Address = len(self.ipv6AddressSet)
        else:
            return False

    def addTCPFlow(self, tcpFlow: TCPFlow):
        """
        每解析到一个 TCP 流就调用本回调
        :param tcpFlow:
        :return:
        """
        self.totalTCPFlowNum += 1

    def checkFileHashes(self, observables: [dict], tcpFlow: TCPFlow):
        """
        检查 observables 中是否包含文件Hash，主要是用于对 HTTP 和 Email 中的 FileHash 进行处理
        :param tcpFlow:
        :param observables:
        :return:
        """
        for observableItem in observables:
            if observableItem["observableType"] == "fileHash":
                self.totalFileNum += 1
                if len(observableItem["observables"]) > 0:
                    hashes = observableItem["value"].split("-")
                    if len(hashes) == 3:
                        fileHash = FileHash()
                        fileHash.md5 = hashes[0]
                        fileHash.sha1 = hashes[1]
                        fileHash.sha256 = hashes[2]
                        self.saveFileHash(fileHash, tcpFlow, self.getBriefObservable(observableItem, False))
                    else:
                        print("HTTP fileHash create error: value = ", observableItem["value"])

    def addEmail(self, mail: Mail, tcpFlow: TCPFlow, observables: [dict]):
        """
        每解析到一个 Email，就调用本方法
        :param observables:
        :param mail:
        :param tcpFlow:
        :return:
        """
        self.totalEmailNum += 1
        self.emailAddressSet.add(mail.From)
        self.emailAddressSet.add(mail.To)
        self.totalEmailAddress = len(self.emailAddressSet)
        self.totalFileNum += len(mail.files)
        self.checkFileHashes(observables, tcpFlow)
        observables = self.addObservable(observables)
        if len(observables) <= 0:
            return
        emailRecord = EmailRecord()
        emailRecord.initByMailAndObservable(mail)
        emailRecord.setTCP(tcpFlow)
        emailRecord.setObservables(observables)
        self.emailRecords.append(emailRecord)

    def addDNSRecord(self, item: DNSItem, observables: [dict]):
        """
        每解析到一个 DNSRecord，就调动本方法
        :param observables:
        :param item:
        :return:
        """
        self.totalDomainNum += 1
        observables = self.addObservable(observables)
        if len(observables) <= 0:
            return
        dnsRecord = DNSRecord()
        dnsRecord.initByDNSItem(item)
        dnsRecord.setObservables(observables)
        self.domainRecords.append(dnsRecord)

    def addHttp(self, httpData: HttpData, tcpFlow: TCPFlow, observables: [dict]):
        """
        每解析到一个 HttpData，就调用本方法
        :param observables:
        [
            {
                "observables":[

                ],
                "value":"a1621.g.akamai.net",
                "observableType": "domain"
            },
            {
                "observables":[

                ],
                "value":"23.15.7.104",
                "observableType": "address"
            }
        ]
        :param httpData:
        :param tcpFlow:
        :return:
        """
        self.totalHTTPNum += 1
        if httpData.getDomain() != "":
            self.totalDomainNum += 1
        self.checkFileHashes(observables, tcpFlow)
        observables = self.addObservable(observables)
        if len(observables) <= 0:
            return
        httpRecord = HTTPRecord(httpData)
        httpRecord.setTCP(tcpFlow)
        httpRecord.setObservables(observables)
        self.httpRecords.append(httpRecord)

    def saveFileHash(self, fileHash: FileHash, tcpFlow: TCPFlow, observables: [dict], filename=""):
        fileHashRecord = FileHashRecord()
        fileHashRecord.fileHash = fileHash
        fileHashRecord.filename = filename
        fileHashRecord.setTCP(tcpFlow)
        fileHashRecord.setObservables(observables)
        self.fileHashRecords.append(fileHashRecord)

    def addFileHash(self, fileHash: FileHash, tcpFlow: TCPFlow, observables: [dict], filename=""):
        """
        每解析到一个 FTP 文件，就调用本方法
        :param filename:
        :param fileHash:
        :param observables:
        :param tcpFlow:
        :return:
        """
        self.totalFTPNum += 1
        self.totalFileNum += 1
        observables = self.addObservable(observables)
        if len(observables) <= 0:
            return
        self.saveFileHash(fileHash, tcpFlow, observables, filename)

    def getBriefObservable(self, observableItem: dict, needCount=True):
        """
        获取 Observable 的简略表示
        {
          "observableId": "xxx",
          "observableType: "ipv4-addr", // ipv4-addr, ipv6-addr, e-mail, domain, uri, fileHash
        }
        :param needCount:
        :param observableItem:      // 是否需要统计
        :return:
        """
        myType, res = "", []
        if len(observableItem["observables"]) == 0:
            return res
        if observableItem["observableType"] == "address":
            observable = observableItem["observables"][0]
            if observable['type'] == 'ipv4-addr':
                # ipv4
                myType = "ipv4-addr"
                if needCount:
                    self.totalMatchIpAddress += 1
            elif observable['type'] == 'ipv6-addr':
                # ipv6
                myType = "ipv6-addr"
                if needCount:
                    self.totalMatchIpv6Address += 1
            elif observable['type'] == 'e-mail':
                # email
                myType = "e-mail"
                if needCount:
                    self.totalMatchEmailNum += 1
        elif observableItem["observableType"] == "domain":
            # domain
            myType = 'domain'
            if needCount:
                self.totalMatchDomain += 1
        elif observableItem["observableType"] == "uri":
            # uri
            myType = 'uri'
            if needCount:
                self.totalMatchUri += 1
        elif observableItem["observableType"] == "fileHash":
            # fileHash
            myType = 'fileHash'
            if needCount:
                self.totalMatchFileHash += 1
        for observable in observableItem["observables"]:
            self.observablesMap[observable["stix_id"]] = observable
            res.append({
                "observableId": observable["stix_id"],
                "observableType": myType
            })
        return res

    def addObservable(self, observables: [dict]):
        """
        统计收到的 Observable 信息
        :param observables:
        [
            {
                "observables":[

                ],
                "value":"a1621.g.akamai.net",
                "observableType": "domain"
            },
            {
                "observables":[

                ],
                "value":"23.15.7.104",
                "observableType": "address"
            }
        ]
        :return:
        """
        res = []
        for observableItem in observables:
            res += self.getBriefObservable(observableItem, True)
        return res

    @staticmethod
    def _getOccurTimeObservableIdObservableUploadTime(timestamp: float, observable: dict):
        return datetime.datetime.fromtimestamp(timestamp).strftime(
            '%Y-%m-%d %H:%M:%S'), observable["stix_id"], datetime.datetime.fromtimestamp(
            observable["create_time"]).strftime(
            '%Y-%m-%d %H:%M:%S')

    class ObservableCSVStatisticsHelper:
        """
        一个对 Observable 出现频次进行静态统计的辅助类
        """
        def __init__(self, fileName):
            self.file = open(fileName, "w")
            self.csvWriter = csv.writer(self.file)
            self.statisticsDict = {}

        def addObservable(self, observable: dict):
            if observable["value"] not in self.statisticsDict:
                self.statisticsDict[observable["value"]] = {
                    observable["stix_id"]: 0
                }
            if observable["stix_id"] not in self.statisticsDict[observable["value"]]:
                self.statisticsDict[observable["value"]][observable["stix_id"]] = 0
            self.statisticsDict[observable["value"]][observable["stix_id"]] += 1

        def addFileHashObservable(self, value: str, observable: dict):
            # print(value, " => ", observable)
            if value not in self.statisticsDict:
                self.statisticsDict[value] = {
                    observable["stix_id"]: 0
                }
            if observable["stix_id"] not in self.statisticsDict[value]:
                self.statisticsDict[value][observable["stix_id"]] = 0
            self.statisticsDict[value][observable["stix_id"]] += 1

        def getCount(self, value: str):
            """
            根据 value 获取其统计值
            :param value:
            :return:
            """
            if value not in self.statisticsDict:
                return 0
            keys = list(self.statisticsDict[value].keys())
            if len(keys) > 0:
                return self.statisticsDict[value][keys[0]]

        def exportCSV(self):
            self.csvWriter.writerow([
                "Serial Number", "value", "observables", "count",
            ])
            seq = 1
            for key, value in self.statisticsDict.items():
                self.statisticsDict.keys()
                for observableId, count in value.items():
                    self.csvWriter.writerow([
                        f"{seq}", key, observableId, f"{count}"
                    ])
                    seq += 1
            self.file.close()

        def exportFileHashCSV(self):
            self.csvWriter.writerow([
                "Serial Number", "md5", "sha1", "sha256", "observables", "count",
            ])
            seq = 1
            for key, value in self.statisticsDict.items():
                hashes = key.split("-")
                if len(hashes) != 3:
                    continue
                for observableId, count in value.items():
                    self.csvWriter.writerow([
                        f"{seq}", hashes[0], hashes[1], hashes[2], observableId, f"{count}"
                    ])
                    seq += 1
            self.file.close()

        def doMerge(self, detailFileName, mergeFileName, startValueIdx, endValueIdx):
            # merge detail and statistics csv table
            with open(detailFileName, "r") as detailFile:
                detailReader = csv.reader(detailFile)
                with open(mergeFileName, "w") as mergeFile:
                    mergeWriter = csv.writer(mergeFile)
                    seq = 0
                    # 先读取掉头部
                    mergeWriter.writerow(["Serial Number", *detailReader.__next__(), "count"])
                    lastV = ""
                    for row in detailReader:
                        v = "-".join(row[startValueIdx:endValueIdx])
                        if lastV != v:
                            lastV = v
                            seq += 1
                        mergeWriter.writerow([
                            str(seq), *row, self.getCount(v)
                        ])

    def _dealIPAddressCSV(self):
        """
        处理导出 IPAddress CSV
        :return:
        """

        def _isIpAddress(ob: dict):
            ob = self.observablesMap[ob["observableId"]]
            return ob['observableType'] == 'address' and (
                    ob['type'] == 'ipv4-addr' or ob['type'] == 'ipv6-addr')

        def _saveRecord(record, timestamp: float, ob: dict):
            if _isIpAddress(ob):
                ob = self.observablesMap[ob["observableId"]]
                ipAddressCSVWriter.writerow([
                    record.srcIP, record.dstIP, ob["value"],
                    *Report._getOccurTimeObservableIdObservableUploadTime(timestamp, ob)
                ])
                ipAddressStatisticsHelper.addObservable(ob)

        ipAddressStatisticsHelper = Report.ObservableCSVStatisticsHelper(
            f"{self.outputCsvDir}/{self.pcapFile}-ip_address_statistics.csv")

        # ip info
        with open(f"{self.outputCsvDir}/{self.pcapFile}-ip_address_detail.csv", "w") as ipAddressCSVFile:
            ipAddressCSVWriter = csv.writer(ipAddressCSVFile)
            ipAddressCSVWriter.writerow(
                ["srcIP", "dstIP", "value", "occur time", "observable id",
                 "observable upload time"]
            )
            # from dnsRecord
            for domainRecord in self.domainRecords:
                for observable in domainRecord.observables:
                    _saveRecord(domainRecord, domainRecord.timestamp, observable)

            # from FTPRecord
            for fileHashRecord in self.fileHashRecords:
                for observable in fileHashRecord.observables:
                    _saveRecord(fileHashRecord, fileHashRecord.endTime, observable)

            # from HTTPRecord
            for httpRecord in self.httpRecords:
                for observable in httpRecord.observables:
                    _saveRecord(httpRecord, httpRecord.endTime, observable)

            # from EmailRecord
            for emailRecord in self.emailRecords:
                for observable in emailRecord.observables:
                    _saveRecord(emailRecord, emailRecord.endTime, observable)

        ipAddressStatisticsHelper.exportCSV()
        ipAddressStatisticsHelper.doMerge(f"{self.outputCsvDir}/{self.pcapFile}-ip_address_detail.csv",
                                          f"{self.outputCsvDir}/{self.pcapFile}-ip_address_merge.csv", 2, 3)

    def _dealEmailAddressCSV(self):
        """
        处理导出 Email Address CSV
        :return:
        """

        emailAddressStatisticsHelper = Report.ObservableCSVStatisticsHelper(
            f"{self.outputCsvDir}/{self.pcapFile}-email_address_statistics.csv")

        with open(f"{self.outputCsvDir}/{self.pcapFile}-email_address_detail.csv", "w") as emailAddressCSVFile:
            emailAddressCSVWriter = csv.writer(emailAddressCSVFile)
            emailAddressCSVWriter.writerow([
                "srcIP", "dstIP", "srcPort", "srcPort", "from", "to", "value", "occur time", "observable id",
                "observable upload time"
            ])
            # from EmailRecord
            for emailRecord in self.emailRecords:
                for observable in emailRecord.observables:
                    observable = self.observablesMap[observable["observableId"]]
                    if observable['observableType'] == 'address' and observable['type'] == 'e-mail':
                        emailAddressCSVWriter.writerow([
                            emailRecord.srcIP, emailRecord.dstIP, emailRecord.srcPort, emailRecord.dstPort,
                            emailRecord.From, emailRecord.To, observable["value"],
                            *Report._getOccurTimeObservableIdObservableUploadTime(emailRecord.endTime, observable)
                        ])
                        emailAddressStatisticsHelper.addObservable(observable)

        emailAddressStatisticsHelper.exportCSV()
        emailAddressStatisticsHelper.doMerge(f"{self.outputCsvDir}/{self.pcapFile}-email_address_detail.csv",
                                             f"{self.outputCsvDir}/{self.pcapFile}-email_address_merge.csv", 6, 7)

    def _dealUriCSV(self):
        """
        处理导出 Uri CSV
        :return:
        """
        uriStatisticsHelper = Report.ObservableCSVStatisticsHelper(
            f"{self.outputCsvDir}/{self.pcapFile}-uri_statistics.csv"
        )

        with open(f"{self.outputCsvDir}/{self.pcapFile}-uri_detail.csv", "w") as uriCSVFile:
            uriCSVWriter = csv.writer(uriCSVFile)
            uriCSVWriter.writerow([
                "srcIP", "dstIP", "srcPort", "dstPort", "value", "occur time", "observable id",
                "observable upload time"
            ])
            for httpRecord in self.httpRecords:
                for observable in httpRecord.observables:
                    observable = self.observablesMap[observable["observableId"]]
                    if observable['observableType'] == 'uri':
                        uriCSVWriter.writerow([
                            httpRecord.srcIP, httpRecord.dstIP, httpRecord.srcPort, httpRecord.dstPort,
                            observable["value"],
                            *Report._getOccurTimeObservableIdObservableUploadTime(httpRecord.endTime, observable)
                        ])
                        uriStatisticsHelper.addObservable(observable)
        uriStatisticsHelper.exportCSV()
        uriStatisticsHelper.doMerge(f"{self.outputCsvDir}/{self.pcapFile}-uri_detail.csv",
                                    f"{self.outputCsvDir}/{self.pcapFile}-uri_merge.csv", 4, 5)

    def _dealFileHashCSV(self):
        """
        处理导出 File hash CSV
        :return:
        """
        fileHashStatisticsHelper = Report.ObservableCSVStatisticsHelper(
            f"{self.outputCsvDir}/{self.pcapFile}-file_hash_statistics.csv"
        )

        with open(f"{self.outputCsvDir}/{self.pcapFile}-file_hash_detail.csv", "w") as fileHashCSVFile:
            fileHashCSVWriter = csv.writer(fileHashCSVFile)
            fileHashCSVWriter.writerow([
                "srcIP", "dstIP", "srcPort", "dstPort", "md5", "sha1", "sha256", "file name", "occur time",
                "observable id",
                "observable upload time"
            ])
            for fileHashRecord in self.fileHashRecords:
                value = fileHashRecord.getValue()
                for observable in fileHashRecord.observables:
                    observable = self.observablesMap[observable["observableId"]]
                    if observable['observableType'] == 'fileHash':
                        fileHashCSVWriter.writerow([
                            fileHashRecord.srcIP, fileHashRecord.dstIP, fileHashRecord.srcPort, fileHashRecord.dstPort,
                            observable["md5"], observable["sha1"], observable["sha256"], fileHashRecord.filename,
                            *Report._getOccurTimeObservableIdObservableUploadTime(fileHashRecord.endTime, observable)
                        ])
                        fileHashStatisticsHelper.addFileHashObservable(value, observable)
        fileHashStatisticsHelper.exportFileHashCSV()
        fileHashStatisticsHelper.doMerge(f"{self.outputCsvDir}/{self.pcapFile}-file_hash_detail.csv",
                                         f"{self.outputCsvDir}/{self.pcapFile}-file_hash_merge.csv", 4, 7)

    def _dealDomainCSV(self):
        """
        处理导出 Domain CSV
        :return:
        """
        domainStatisticsHelper = Report.ObservableCSVStatisticsHelper(
            f"{self.outputCsvDir}/{self.pcapFile}-domain_statistics.csv"
        )

        with open(f"{self.outputCsvDir}/{self.pcapFile}-domain_detail.csv", "w") as domainCSVFile:
            domainCSVWriter = csv.writer(domainCSVFile)
            domainCSVWriter.writerow([
                "srcIP", "dstIP", "srcPort", "dstPort", "value", "occur time", "observable id",
                "observable upload time"
            ])
            # from Domain record
            for domainRecord in self.domainRecords:
                for observable in domainRecord.observables:
                    observable = self.observablesMap[observable["observableId"]]
                    if observable["observableType"] == 'domain':
                        domainCSVWriter.writerow([
                            domainRecord.srcIP, domainRecord.dstIP, str(53), "",
                            observable["value"],
                            *Report._getOccurTimeObservableIdObservableUploadTime(domainRecord.timestamp, observable)
                        ])
                        domainStatisticsHelper.addObservable(observable)

            # from http record
            for httpRecord in self.httpRecords:
                for observable in domainRecord.observables:
                    observable = self.observablesMap[observable["observableId"]]
                    if observable["observableType"] == 'domain':
                        domainCSVWriter.writerow([
                            httpRecord.srcIP, httpRecord.dstIP, httpRecord.srcPort, httpRecord.dstPort,
                            observable["value"],
                            *Report._getOccurTimeObservableIdObservableUploadTime(httpRecord.endTime, observable)
                        ])
                        domainStatisticsHelper.addObservable(observable)
        domainStatisticsHelper.exportCSV()
        domainStatisticsHelper.doMerge(f"{self.outputCsvDir}/{self.pcapFile}-domain_detail.csv",
                                       f"{self.outputCsvDir}/{self.pcapFile}-domain_merge.csv", 4, 5)

    def end(self):
        """
        报告结束
        self.emailRecords = []  # contain all email which some observable match it
        self.fileHashRecords = []  # contain all FTP session which some observable match it
        self.httpRecords = []  # contain all HTTP session which some observable match it
        self.domainRecords = []  #
        :return:
        """
        if self.isDumpCsv:
            if not os.path.exists(self.outputCsvDir):
                os.makedirs(self.outputCsvDir)
            self._dealIPAddressCSV()
            self._dealEmailAddressCSV()
            self._dealUriCSV()
            self._dealFileHashCSV()
            self._dealDomainCSV()



from .Extractor import Extractor
from dpkt import ethernet, ip, icmp
from dpkt.utils import inet_to_str


class ICMPExtractor(Extractor):
    def __init__(self, valueCallback):
        super().__init__(valueCallback)

    def addPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        # 过滤一下，我们只处理ICMP包
        if not (isinstance(ethPacket.data, ip.IP) and isinstance(ethPacket.data.data, icmp.ICMP)):
            return
        ipPacket = ethPacket.data
        icmpPacket = ipPacket.data
        self.valueCallback({
            "src_ip": inet_to_str(ipPacket.src),
            "dst_ip": inet_to_str(ipPacket.dst),
            "icmp_type": icmpPacket.type,
            "icmp_code": icmpPacket.code
        })

    def done(self):
        pass



from io import BytesIO
from contextlib import closing
from .mail import Mail
from email.parser import Parser


class SMTPParser:
    """
    RFC: https://datatracker.ietf.org/doc/html/rfc5321

    SMTP 解析器，参考：https://www.rfc-editor.org/rfc/rfc5321.html#section-4.1
    EHLO => 向SMTP服务器提交本地客户端信息，用来获取 SMTP 邮件服务器的一些描述信息， eg.: ehlo Zoey.local
    HELO => 向SMTP服务器提交本地客户端信息，用来获取 SMTP 邮件服务器的一些描述信息， eg.: ehlo Zoey.local
    MAIL => 启动一个邮件事务（包含发件人的地址），eg.: mail FROM:<123zoeyyy@163.com>
    RCPT => 用这个命令标识单个邮件收件人，eg.: rcpt TO:<yzou10@uoguelph.ca>
    DATA => 标识邮件数据的起始，eg.: data
    RSET => 忽略
    VRFY => 忽略
    AUTH => 发送验证信息，忽略
    EXPN => 忽略
    HELP => client向server请求帮助信息，忽略
    NOOP => 忽略
    QUIT => 结束传输通道
    """

    def __init__(self):
        self.commandMap = {
            b"EHLO": self.dealEHLO,
            b"HELO": self.dealHELO,
            b"MAIL": self.dealMAIL,
            b"RCPT": self.dealRCPT,
            b"DATA": self.dealDATA,
            b"RSET": self.ignore,
            b"VRFY": self.ignore,
            b"AUTH": self.ignore,
            b"EXPN": self.ignore,
            b"HELP": self.ignore,
            b"NOOP": self.ignore,
            b"QUIT": self.dealQUIT
        }

    def dealEHLO(self, mail: Mail, msg: bytes):
        pass

    def dealHELO(self, mail: Mail, msg: bytes):
        pass

    def dealMAIL(self, mail: Mail, msg: bytes):
        pass

    def dealRCPT(self, mail: Mail, msg: bytes):
        pass

    def dealDATA(self, mail: Mail, msg: bytes):
        pass

    def dealQUIT(self, mail: Mail, msg: bytes):
        """
        包头：https://help.aliyun.com/document_detail/51584.html
        :param mail:
        :param msg:
        :return:
        """
        if msg.endswith(b"\r\n.\r\nquit\r\n") or msg.endswith(b"\r\n.\r\nQuit\r\n") or \
                msg.endswith(b"\r\n.\r\nQUIT\r\n"):
            msg = msg[:len(msg) - 11]
        msg = Parser().parsestr(msg.decode('utf-8'))
        mail.parse(msg)

    def ignore(self, mail: Mail, msg: bytes):
        pass

    def parse(self, data: bytes) -> [Mail]:
        """
        传入单向的 TCP 流，client => server
        :param data:
        :return:
        """
        mails = []
        with closing(BytesIO(data)) as data:
            line = data.readline()
            msg, mail, dataBegin = bytes(), Mail(), False
            while line not in [b'']:
                if len(line) < 4:
                    msg += line
                    line = data.readline()
                    continue
                command = line[:4].upper() 
                if command in self.commandMap:
                    if command == b"RCPT":
                        mail = Mail()
                        mails.append(mail)
                    if command == b"DATA":
                        msg, dataBegin = b"".join(data.readlines()), True
                        self.dealQUIT(mail, msg)
                        break
                    self.commandMap[command](mail, msg)

                    msg = bytes()
                else:
                    msg += line

                line = data.readline()
        return mails



from abc import ABCMeta, abstractmethod
from dpkt import ethernet


class Extractor(object):
    """
    提取器的基类，定义了所有提取器通用的功能
    """
    __metaclass__ = ABCMeta  # 指定这是一个抽象类

    def __init__(self, valueCallback):
        self.valueCallback = valueCallback

    @abstractmethod
    def addPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        pass

    @abstractmethod
    def done(self):
        pass



from .Extractor import Extractor
from dpkt import ethernet, dns, ip, udp, UnpackError
from dpkt.utils import inet_to_str
from datetime import datetime


class DNSItem:
    domain = ""
    domain_type = ""
    value = ""
    timestamp = 0

    def __init__(self, values):
        self.domain = values[0]
        self.domain_type = values[1]
        self.value = values[2]
        self.timestamp = values[3]
        self.ethPacket = values[4]

    def toDomainRecord(self):
        return {
            "domain": self.domain,
            "domain_type": self.domain_type,
            "value": self.value,
            "timestamp": datetime.fromtimestamp(self.timestamp)
        }


class DNSExtractor(Extractor):

    def __init__(self, valueCallback):
        super().__init__(valueCallback)

    def addPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        # DNS 采用UDP通信，不是UDP包忽略
        if not (isinstance(ethPacket.data, ip.IP) and isinstance(ethPacket.data.data, udp.UDP)):
            return
        ipPacket = ethPacket.data
        udpPacket = ipPacket.data
        if udpPacket.sport == 53:
            # 源端口是53，表示是域名服务器返回的 Response
            try:
                dnsPacket = dns.DNS(udpPacket.data)
            except UnpackError:
                return
            # 下面几种section的含义，参考这个链接：http://www.ruanyifeng.com/blog/2016/06/dns.html
            for rr in dnsPacket.qd:
                # question section
                pass
            for rr in dnsPacket.an:
                # answer section
                self.decodeDNSResponse(rr, timestamp, ethPacket)
            for rr in dnsPacket.ns:
                # authority section
                pass
            for rr in dnsPacket.ar:
                # addition section
                pass

    def decodeDNSResponse(self, rr: dns.DNS.RR, timestamp: float, ethPacket: ethernet.Ethernet):
        if rr.type == dns.DNS_A:
            # A 记录 => ipv4
            self.valueCallback(DNSItem([rr.name, "A", inet_to_str(rr.rdata), timestamp, ethPacket]))
        elif rr.type == dns.DNS_AAAA:
            # AAAA 记录 => ipv6
            self.valueCallback(DNSItem([rr.name, "AAAA", inet_to_str(rr.rdata), timestamp, ethPacket]))

    def done(self):
        pass



import re

from .mail import Mail
from io import BytesIO
from contextlib import closing
from email.parser import Parser


class IMAPParser:
    """
    RFC: https://datatracker.ietf.org/doc/html/rfc3501
    """
    def parse(self, commandBytes: bytes, responseBytes: bytes) -> [Mail]:
        """
        解析 IMAP TCP流，从其中解析出邮件
        :param commandBytes:
        :param responseBytes:
        :return:
        """
        emails = []
        with closing(BytesIO(commandBytes)) as data:
            commands = data.readlines()
        if len(commands) < 0:
            return []
        commandIndex = 0
        start = 0
        responses = []
        currentCommandId, isFetch = commands[commandIndex][:5], commands[commandIndex][6:].upper().startswith(b"FETCH")
        with closing(BytesIO(responseBytes)) as data:
            lines = data.readlines()
        for idx, line in enumerate(lines):
            if line.startswith(currentCommandId):
                if isFetch:
                    responses.append(b''.join(lines[start:idx]))
                start = idx + 1
                commandIndex += 1
                if commandIndex < len(commands):
                    currentCommandId, isFetch = commands[commandIndex][:5], commands[commandIndex][
                                                                            6:].upper().startswith(b"FETCH") or \
                                                commands[commandIndex][6:].upper().startswith(b"UID FETCH")
                else:
                    break

        for response in responses:
            # 在这里处理 fetch 的结果
            with closing(BytesIO(response)) as data:
                result = re.search(rb'\* [0-9]+ FETCH(.*?){(.*?)}\r\n', data.readline())
                while result is not None:
                    size = int(result.group(2))
                    tmp = data.read(size)
                    # 忽略掉结束标识
                    data.read(3)
                    mail = Mail()
                    msg = Parser().parsestr(tmp.decode('utf-8'))
                    if msg.get("From") is not None and msg.get("To") is not None:
                        mail.parse(msg)
                        emails.append(mail)
                    result = re.search(rb'\* [0-9]+ FETCH(.*?){(.*?)}\r\n', data.readline())
        # print(responses[7], len(responses[7]))
        return emails



# -*- encoding: UTF-8 -*-
from setuptools import setup, find_packages

VERSION = '0.1.7'

setup(name='cctxpa',
      version=VERSION,
      description="A command line tool for CCTX to analyser pcap and compare with CCTX's observables",
      long_description='',
      classifiers=[],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='python cctx pcap cctxpa terminal',
      author='zoeyyy',
      author_email='yzou10@uoguelph.ca',
      url='https://github.com/zoeyyyzou/cctx-pcap-analyser',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          'dpkt',
          'progress',
          'requests'
      ],
      entry_points={
          'console_scripts': [
              'cctxpa = cctxpa.cctxpa:main'
          ]
      },
      )



from abc import ABCMeta, abstractmethod
from dpkt import ip, ethernet, ip6, tcp, udp
from dpkt.utils import inet_to_str, mac_to_str


class FlowBase(metaclass=ABCMeta):
    """
    定义网络流（TCP流和UDP流的通用属性）
    """
    startTime = 0  # 流起始时间
    lastTime = 0  # 最后一次更新的时间
    lastForwardTime = 0  # 最后一次抓取到正向包的时间
    lastReverseTime = 0  # 最后一次抓取到反向包的时间
    # srcMacAddr = ""  # 源mac地址
    # dstMacAddr = ""  # 目的mac地址
    srcIP = ""  # 源IP地址
    dstIP = ""  # 目的IP地址
    srcPort = 0  # 源端口
    dstPort = 0  # 目的端口
    category = ""  # 地址类型 "ipv4-addr" / "ipv6-addr"
    protocol = 0  # 协议

    forwardStr = ""  # 正向流字符串表示，例如 6-192.168.1.4:13853-192.168.1.5:80
    reverseStr = ""  # 反向流字符串表示，例如 6-192.168.1.5:80-192.168.1.4:13853

    # 统计值
    totalCount = 0  # 总的网络包的数量
    totalForwardCount = 0  # 总的正向网络包的数量
    totalReverseCount = 0  # 总的反向网络包的数量
    totalPayloadBytes = 0  # 总的 payload 字节数
    totalForwardBytes = 0  # 总的正向的 payload 字节数
    totalReverseBytes = 0  # 总的反向的 payload 字节数

    def __init__(self, ethPacket: ethernet.Ethernet, timestamp: float):
        if not FlowBase.canBeMarkAsFlow(ethPacket):
            # 如果不是TCP或者UDP包，则忽略
            return
        self.startTime = timestamp
        # self.srcMacAddr = mac_to_str(ethPacket.src)
        # self.dstMacAddr = mac_to_str(ethPacket.dst)

        # 开始提取IP信息
        ipPacket = ethPacket.data
        self.srcIP = inet_to_str(ipPacket.src)
        self.dstIP = inet_to_str(ipPacket.dst)

        if isinstance(ipPacket, ip.IP):
            self.protocol = ipPacket.p
            self.category = "ipv4-addr"
        else:
            self.protocol = ipPacket.nxt
            self.category = "ipv6-addr"

        if self.protocol == ip.IP_PROTO_TCP and isinstance(ipPacket.data, tcp.TCP):
            # TCP 包
            tcpPacket = ipPacket.data
            self.srcPort = tcpPacket.sport
            self.dstPort = tcpPacket.dport
        elif self.protocol == ip.IP_PROTO_UDP and isinstance(ipPacket.data, udp.UDP):
            # UDP包
            udpPacket = ipPacket.data
            self.srcPort = udpPacket.sport
            self.dstPort = udpPacket.dport

        # 保存当前网络流的正反方向字符串表示
        self.forwardStr = f"{FlowBase.getProtocol(ethPacket)}-{self.srcIP}:{self.srcPort}-{self.dstIP}:{self.dstPort}"
        self.reverseStr = f"{FlowBase.getProtocol(ethPacket)}-{self.dstIP}:{self.dstPort}-{self.srcIP}:{self.srcPort}"

        # 调用 addPacket 添加第一个这个网络流的第一个网络包
        self.addForwardPacket(ethPacket, timestamp)

    def _addForwardPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        """
        在此处对 TCP 和 UDP 流的一些正向流的共性特征做提取和统计
        :param ethPacket:
        :param timestamp:
        :return:
        """
        self.lastTime = timestamp
        self.lastForwardTime = timestamp

        # 静态统计相关
        self.totalCount += 1
        self.totalForwardCount += 1
        if ethPacket.data.data.data:
            self.totalPayloadBytes += len(ethPacket.data.data.data)
            self.totalForwardBytes += len(ethPacket.data.data.data)

    def _addReversePacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        """
        在此处对 TCP 和 UDP 流的一些反向流的共性特征做提取和统计
        :param ethPacket:
        :param timestamp:
        :return:
        """
        self.lastTime = timestamp
        self.lastReverseTime = timestamp

        # 静态统计相关
        self.totalCount += 1
        self.totalReverseCount += 1
        if ethPacket.data.data.data:
            self.totalPayloadBytes += len(ethPacket.data.data.data)
            self.totalReverseBytes += len(ethPacket.data.data.data)

    @abstractmethod
    def addForwardPacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        pass

    @abstractmethod
    def addReversePacket(self, ethPacket: ethernet.Ethernet, timestamp: float):
        pass

    @abstractmethod
    def getAllForwardBytes(self) -> bytes:
        """
        获取正向流的所有字节数据
        :return:
        """
        pass

    @abstractmethod
    def getAllReverseBytes(self) -> bytes:
        """
        获取反向流的所有字节数据
        :return:
        """
        pass

    @staticmethod
    def getProtocol(ethPacket: ethernet.Ethernet):
        """
        提取 IPv6 和 IPv4 包中的协议字段
        :param ethPacket:
        :return:
        """
        ipPacket = ethPacket.data
        if isinstance(ipPacket, ip.IP):
            return ipPacket.p
        elif isinstance(ipPacket, ip6.IP6):
            return ipPacket.nxt
        return 0

    @staticmethod
    def encodeToStr(ethPacket: ethernet.Ethernet):
        """
        将 TCP/UDP 包 编码成字符串表示，格式如下：
        <协议>-<源ip>:<源端口>-<目的ip>:<目的端口>
        例如：
            6-192.168.1.4:13853-192.168.1.5:80
        :param ethPacket:
        :return:
        """
        # 开始提取IP信息
        ipPacket = ethPacket.data
        srcIP = inet_to_str(ipPacket.src)
        dstIP = inet_to_str(ipPacket.dst)
        return f"{FlowBase.getProtocol(ethPacket)}-{srcIP}:{ipPacket.data.sport}-{dstIP}:{ipPacket.data.dport}"

    @staticmethod
    def encodeToReverseStr(ethPacket: ethernet.Ethernet):
        # 开始提取IP信息
        ipPacket = ethPacket.data
        srcIP = inet_to_str(ipPacket.src)
        dstIP = inet_to_str(ipPacket.dst)
        return f"{FlowBase.getProtocol(ethPacket)}-{dstIP}:{ipPacket.data.dport}-{srcIP}:{ipPacket.data.sport}"

    @staticmethod
    def canBeMarkAsFlow(ethPacket: ethernet.Ethernet):
        """
        判断收到的一个以太网包里面存放的是不是 TCP/UDP 包
        1. 如果是 TCP/UDP 包，则可以将其标识为一个网络流;
        2. 如果不是，则不能标记为网络流
        :param ethPacket:
        :return:
        """
        if isinstance(ethPacket.data, ip.IP):
            # 如果是 IPv4 包
            return (ethPacket.data.p == ip.IP_PROTO_TCP and isinstance(ethPacket.data.data, tcp.TCP)) \
                   or (ethPacket.data.p == ip.IP_PROTO_UDP and isinstance(ethPacket.data.data, udp.UDP))
        elif isinstance(ethPacket.data, ip6.IP6):
            return (ethPacket.data.nxt == ip.IP_PROTO_TCP and isinstance(ethPacket.data.data, tcp.TCP)) \
                   or (ethPacket.data.nxt == ip.IP_PROTO_UDP and isinstance(ethPacket.data.data, udp.UDP))
        else:
            return False



import argparse
import datetime
import json
import os
import sys
import dpkt
import requests

from .pcap_extractor.MyEncoder import MyEncoder
from .pcap_extractor.FlowExtractor import FlowExtractor
from .pcap_extractor.TCPFlow import TCPFlow
from .pcap_extractor.mail import Mail
from .pcap_extractor.HTTPParser import HTTPParser, HttpData
from .pcap_extractor.SMTPParser import SMTPParser
from .pcap_extractor.POP3Parser import POP3Parser
from .pcap_extractor.IMAPParser import IMAPParser
from .pcap_extractor.DNSExtractor import DNSExtractor, DNSItem
from .report import Report, FileHash
from progress.bar import Bar
from rich.console import Console
from rich.table import Table
from rich.markdown import Markdown
from rich.columns import Columns
from rich.panel import Panel


class CCTXPcapAnalyser:
    def __init__(self, args):
        self.host = args.host                   # cctx-pcap-analyser 服务器地址（可以是ip或者域名，默认是 127.0.0.1）
        self.path = args.path                   # cctx-pcap-analyser server 的API相对路径
        self.port = args.port                   # cctx-pcap-analyser server 的端口，默认为 5000
        self.https = args.https                 # 是否使用 https，默认为 True
        self.username = args.username           # 用来登录或者注册 cctx-pcap-analyser server 的用户名
        self.password = args.password           # 用来登录或者注册 cctx-pcap-analyser server 的密码
        self.inputFile = args.pcapfile          # 要解析的 pcap 文件
        self.outputFile = args.outputfile       # 指定输出的 json 格式的report的名字，默认为 report.json
        self.progress = args.progress           # 是否输出提取进度条，默认为 True
        if self.port == -1:
            # 不指定端口，使用默认的80或者443
            if self.https:
                self.port = 443
            else:
                self.port = 80

        # 实例化一个 FlowExtractor，用来从 pcap 文件里面提取一个个 TCP 流
        # 并且每提取到一个 TCP 流，就会调用 self.dealStream 这个回调来处理它
        self.flowExtractor = FlowExtractor(valueCallback=self.dealStream)

        # 实例化一个 DNSExtractor，用来从 pcap 文件里面提取 DNS 查询记录
        # 并且每提取到一个 DNS 查询记录，就会调用 self.dealDNSRecord 这个回调来处理它
        self.dnsExtractor = DNSExtractor(valueCallback=self.dealDNSRecord)

        # 初始化一系列的 Parser，用来从 TCP 流量中解析应用层的数据（包括 HTTP 请求回复，和三种邮件协议的解析）
        self.httpParser = HTTPParser()
        self.smtpParser = SMTPParser()
        self.pop3Parser = POP3Parser()
        self.imapParser = IMAPParser()

        # 实例化 Report，提取过程中的相关记录都会存到该对象中
        self.report = Report()
        self.report.pcapFile = args.pcapfile
        self.report.outputCsvDir = args.outputcsvdirectory

        # 根据传入的参数构造出几个 URL
        # 用来注册的接口的URL => http://127.0.0.1:5000/user/register
        # 用来登录的接口的URL => http://127.0.0.1:5000/user/login
        # 用来查询 Observable 的接口的URL => http://127.0.0.1:5000/user/queryObservables
        self.url = f'{"https" if self.https else "http"}://{self.host}:{self.port}{self.path}'
        self.registerUrl = f'{self.url}/user/register'
        self.loginUrl = f'{self.url}/user/login'
        self.queryUrl = f'{self.url}/user/queryObservables'

        # 用来保存登录时返回的 token，因为查询 Observable 的接口需要登录后携带 token 才能请求
        self.token = ""
        # 用来保存 refreshToken （暂时没用到）
        # refreshToken 的有效时间一般比 token 长很多，当 token 过期时，可以使用 refreshToken 来申请更新 token
        self.refreshToken = ""

        self.report.begin()

        # 一个字典，用来缓存查询结果（主要用来查询加速）
        # {
        #   "google.com": {
        #                       "observables": []
        #                 },
        #   "217.31.42.56": {
        #                        "observables": []
        #                  },
        # }
        #
        # 1. 比如之前查询过 "abv.com" 这个域名，服务器返回 observables 为空表示没有 observable 和 这个域名匹配，使用 queryCache 保存
        #    这个结果，那么下一次再次查询这个域名的时候，直接从 queryCache 中就能查询到为空，不需要向服务器发起请求了，可以起到查询加速的目的。
        self.queryCache = {

        }

    def doAuthGet(self, url: str, params: dict):
        """
        携带 Token 发起一个 Get请求
        1. 因为有些 API 接口需要登录之后携带token才能访问
        :param url:
        :param params:
        :return:
        """
        r = requests.get(url, params=params, headers={
            "Authorization": f"Bearer {self.token}"         # 在请求头中携带 token
        })
        # print(r.json())
        if r.status_code == 401:
            # 如果权限验证过期，则表示token失效了，只要重新登录刷新token，并再次尝试发起刚才失败的请求即可
            # 通过这种方式，即便 token 的有效期很短，使用命令行工具的用户也感觉不到
            self.doLogin()
            return self.doAuthGet(url, params)
        elif r.status_code != 200:
            return {
                "code": 500,
                "msg": r.reason,
                "data": ""
            }
        responseData = r.json()
        return responseData

    def doAuthPost(self, url: str, jsonData: dict):
        """
        携带 Token 发起一个 post 请求
        1. 因为有些 API 接口需要登录之后携带token才能访问
        :param url:
        :param jsonData:
        :return:
        """
        r = requests.post(url, json=jsonData, headers={
            "Authorization": f"Bearer {self.token}"  # 在请求头中携带 token
        })
        responseData = r.json()
        if responseData == 401:
            # 如果权限验证过期，则表示token失效了，只要重新登录刷新token，并再次尝试发起刚才失败的请求即可
            # 通过这种方式，即便 token 的有效期很短，使用命令行工具的用户也感觉不到
            self.doLogin()
            return self.doAuthPost(url, jsonData)
        return responseData

    @staticmethod
    def buildQueryDomain(domain: str) -> dict:
        """
        构造一个查询 Domain 表的子任务
        :param domain:
        :return:
        """
        return {
            'queryType': 'domain',
            'value': domain
        }

    @staticmethod
    def buildQueryAddress(address: str) -> dict:
        """
        构造一个查询 Address 表的子任务
        :param address:
        :return:
        """
        return {
            'queryType': 'address',
            'value': address
        }

    @staticmethod
    def buildQueryUri(uri: str) -> dict:
        """
        构造一个查询 URI 表的子任务
        :param uri:
        :return:
        """
        return {
            'queryType': 'uri',
            'value': uri
        }

    @staticmethod
    def buildQueryFileHash(fileHash: FileHash) -> dict:
        """
        构造一个查询 FileHash 表的子任务
        :param fileHash:
        :return:
        """
        return {
            'queryType': 'fileHash',
            'md5': fileHash.md5,
            'sha1': fileHash.sha1,
            'sha256': fileHash.sha256
        }

    def doQuery(self, queryItems: [dict]) -> [dict]:
        """
        执行一组查询任务。执行下列步骤
        1. 遍历每一个查询任务，对每一个查询子任务，检查本地的 queryCache 是否有缓存对应value的结果，
           如果有直接使用，如果没有则将该子任务加到 remainQueryItems 列表当中；
        2. 然后检查 remainQueryItems 是否为空，不为空则向去服务发起一个查询请求；
        3. 然后合并步骤1和步骤2的结果，并返回
        :param queryItems:
                [
                    {
                        "queryType": "address",
                        "value": "192.168.21.45"
                    },
                    {
                        "queryType": "address",
                        "value": "21.3.23.45"
                    },
                    {
                        "queryType": "domain",
                        "value": "asb.com"
                    },
                ]
        :return:
        """
        def getValue(queryItem: dict):
            if queryItem["queryType"] == "fileHash":
                return f'{queryItem["md5"]}-{queryItem["sha1"]}-{queryItem["sha256"]}'
            return queryItem["value"]

        # results => 存储查询对比的结果
        # remainQueryItems => 存储本地缓存没有查到，需要到服务器后台查询的任务
        results, remainQueryItems = [], []
        for queryItem in queryItems:
            value = getValue(queryItem)
            if value in self.queryCache:
                if len(self.queryCache[value]["observables"]) > 0:
                    results.append(self.queryCache[value])
            else:
                remainQueryItems.append(queryItem)

        # 如果所有的查询任务都命中了本地缓存，则直接返回结果就可，不用请求服务器后台
        if len(remainQueryItems) == 0:
            return results

        # 如果还有没有命中缓存的查询任务需要查询，则向服务器后台发送查询请求
        response = self.doAuthGet(self.queryUrl, params={
            'queries': json.dumps(remainQueryItems),
        })
        if response['code'] == 200:
            if response['data'] and len(response['data']) > 0:
                for observable in response['data']:
                    self.queryCache[observable["value"]] = observable
                results += response['data']
        else:
            print(f'Query Observable failed, {response["msg"]}')
        return results

    def dealDNSRecord(self, dnsRecord: DNSItem):
        """
        处理每个提取到的 DNS 解析记录
        :param dnsRecord:
        :return:
        """
        # query DNSItem
        observables = self.doQuery([
            CCTXPcapAnalyser.buildQueryDomain(dnsRecord.domain),
            CCTXPcapAnalyser.buildQueryAddress(dnsRecord.value)
        ])
        self.report.addDNSRecord(dnsRecord, observables)

    def dealMail(self, mail: Mail, tcpFlow: TCPFlow):
        """
        处理每个解析出来的邮件
        :param mail:
        :param tcpFlow:
        :return:
        """
        # query email address, ip address, file hash
        queryTask = [
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.srcIP),
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.dstIP),
            CCTXPcapAnalyser.buildQueryAddress(mail.From),
            CCTXPcapAnalyser.buildQueryAddress(mail.To),
        ]
        for mailFile in mail.files:
            fileHash = FileHash(mailFile.fileData)
            queryTask.append(CCTXPcapAnalyser.buildQueryFileHash(fileHash))
        observables = self.doQuery(queryTask)
        self.report.addEmail(mail, tcpFlow, observables)

    def dealHttpData(self, httpData: HttpData, tcpFlow: TCPFlow):
        """
        处理每个解析出来的 HttpData
        :param httpData:
        :param tcpFlow:
        :return:
        """
        # query url, file hash, ip address, domain
        queryTask = [
            CCTXPcapAnalyser.buildQueryDomain(httpData.getDomain()),
            CCTXPcapAnalyser.buildQueryUri(httpData.getUrl()),
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.srcIP),
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.dstIP),
        ]
        if httpData.fileHashes:
            fileHash = FileHash()
            fileHash.md5 = httpData.fileHashes["md5"]
            fileHash.sha1 = httpData.fileHashes["sha1"]
            fileHash.sha256 = httpData.fileHashes["sha256"]
            queryTask.append(CCTXPcapAnalyser.buildQueryFileHash(fileHash))
        observables = self.doQuery(queryTask)
        self.report.addHttp(httpData, tcpFlow, observables)

    def dealFTP(self, data: bytes, tcpFlow: TCPFlow):
        """
        处理每个 FTP
        :param data:
        :param tcpFlow:
        :return:
        """
        fileHash = FileHash(data)
        queryTask = [
            CCTXPcapAnalyser.buildQueryFileHash(fileHash),
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.srcIP),
            CCTXPcapAnalyser.buildQueryAddress(tcpFlow.dstIP),
        ]
        observables = self.doQuery(queryTask)
        self.report.addFileHash(fileHash, tcpFlow, observables)

    def dealStream(self, tcpFlow: TCPFlow):
        """
        处理每个提取到的TCP流
        :param tcpFlow:
        :return:
        """
        self.report.addTCPFlow(tcpFlow)
        forwardBytes, reverseBytes = tcpFlow.getAllForwardBytes(), tcpFlow.getAllReverseBytes()
        if tcpFlow.dstPort == 21:
            pass
        elif tcpFlow.srcPort == 20 or tcpFlow.dstPort == 20:
            # 处理主动模式 FTP
            data1, data2 = forwardBytes, reverseBytes
            data = data1 if len(data2) == 0 else data2
            if len(data) > 0:
                # md1 = hashlib.md5()
                # md2 = hashlib.md5()
                # md3 = hashlib.md5()
                # with closing(BytesIO(data)) as data:
                #     for line in data.readlines():
                #         md1.update(line)
                #         if line.endswith(b"\r\n"):
                #             md2.update(line[:-2])
                #             md2.update(b'\r')
                #             md3.update(line[:-2])
                #             md3.update(b'\n')
                self.dealFTP(data, tcpFlow)
        elif tcpFlow.dstPort == 143:
            # 处理 IMAP
            for mail in self.imapParser.parse(forwardBytes, reverseBytes):
                self.dealMail(mail, tcpFlow)
        elif tcpFlow.dstPort == 110:
            # 处理 POP3
            for mail in self.pop3Parser.parse(reverseBytes):
                self.dealMail(mail, tcpFlow)
        elif tcpFlow.dstPort == 25:
            # 处理 SMTP
            for mail in self.smtpParser.parse(forwardBytes):
                self.dealMail(mail, tcpFlow)
        elif (len(forwardBytes) == 0 and len(reverseBytes) > 0) or (len(forwardBytes) > 0 and len(reverseBytes) == 0):
            # 处理被动模式 FTP
            # try to cal file hash for FTP passive mode
            if len(forwardBytes) == 0 and len(reverseBytes) > 0:
                self.dealFTP(reverseBytes, tcpFlow)
            else:
                self.dealFTP(forwardBytes, tcpFlow)
        else:
            # parse http
            for httpData in self.httpParser.parse(forwardBytes, reverseBytes):
                self.dealHttpData(httpData, tcpFlow)

    def doRegister(self):
        """
        执行注册
        :return:
        """
        r = requests.post(self.registerUrl, json={
            'username': self.username,
            'password': self.password
        })
        return r.json()

    def doLogin(self):
        """
        执行登录
        :return:
        """
        r = requests.post(self.loginUrl, json={
            'username': self.username,
            'password': self.password
        })
        # print("Do login")
        try:
            data = r.json()
        except Exception as e:
            print(f"Login failed, {r.reason}")
            return False
        if data['code'] != 200:
            print(f"Login failed, {data['msg']}")
            return False
        self.token = data['data']['token']
        self.refreshToken = data['data']['refresh_token']
        return True

    def start(self):
        """
        Start to parse pcap file
        :return:
        """

        # do login first
        if not self.doLogin():
            return

        # 判断是否要展示进度条
        if self.progress:
            bar = Bar('Extract Progress:', max=100)
        else:
            bar = None

        self.report.begin()
        with open(self.inputFile, 'rb') as pcap:
            progress = 0
            totalSize = os.path.getsize(self.inputFile)
            packets = dpkt.pcap.Reader(pcap)
            for ts, buf in packets:
                if self.progress:
                    currentProgress = int(pcap.tell() * 100 / totalSize)
                    if currentProgress != progress:
                        progress = currentProgress
                        bar.next()
                ethPacket = dpkt.ethernet.Ethernet(buf)
                self.report.addPacket(ethPacket, ts)
                self.flowExtractor.addPacket(ethPacket, ts)
                self.dnsExtractor.addPacket(ethPacket, ts)
        if self.progress:
            bar.finish()
        self.report.end()
        self.flowExtractor.done()


def printReport(report: Report, args):
    """
    将对比报告输出到控制台终端
    :param report:
    :return:
    """

    def getPanelContent(key, value: str = None):
        return Panel(f"[b]{key}[/b]\n[yellow]{getattr(report, key) if value is None else value}", expand=True, width=30)

    console = Console()
    console.print(Markdown("# Analyser Report"))
    durationSeconds = report.endTime - report.startTime
    durationMinutes = durationSeconds / 60
    durationSeconds = durationSeconds % 60
    durationHours = durationMinutes / 60
    durationMinutes = durationMinutes % 60
    fsize = os.path.getsize(args.pcapfile)
    renderAbles = [
                      getPanelContent('input pcap file', args.pcapfile),
                      getPanelContent('detail report', f"{args.pcapfile}-{args.outputfile}"),
                      getPanelContent('startTime',
                                      datetime.datetime.fromtimestamp(report.startTime).strftime('%Y-%m-%d %H:%M:%S')),
                      getPanelContent('endTime',
                                      datetime.datetime.fromtimestamp(report.endTime).strftime('%Y-%m-%d %H:%M:%S')),
                      getPanelContent('duration',
                                      f"{int(durationHours)} h : {int(durationMinutes)} m : {int(durationSeconds)} s"),
                      getPanelContent('fileSize', f"{round(fsize/float(1024*1024), 2)} MB")

                      # getPanelContent('duration', f"{report.endTime - report.startTime}"),
                  ] + [getPanelContent(item) for item in
                       ['totalPacket', 'totalIPAddress', 'totalIPv6Address', 'totalEmailAddress', 'totalIPPacket',
                        'totalIPv6Packet',
                        'totalTCPFlowNum', 'totalHTTPNum', 'totalFTPNum', 'totalEmailNum', 'totalFileNum',
                        'totalDomainNum',
                        'totalMatchIpAddress', 'totalMatchIpv6Address', 'totalMatchEmailNum', 'totalMatchDomain',
                        'totalMatchFileHash', 'totalMatchUri']]
    columns = Columns(renderAbles, equal=True, expand=True)
    console.print(columns)

    console.print(Markdown("## CCTX Observable Match Results"))
    table = Table(expand=True)
    table.add_column("Extracted Feature", justify='center', style='cyan')
    table.add_column("Extracted Feature Count", justify='center', style='cyan')
    table.add_column("Matched Observable Count", justify='center', style='magenta')
    table.add_column("Matching Percentage", justify='center', style='green')
    table.add_row('IPv4 address', f"{report.totalIPAddress}", f"{report.totalMatchIpAddress}",
                  f"{0 if report.totalIPAddress == 0 else round(report.totalMatchIpAddress * 100.0 / report.totalIPAddress, 2)}%")
    table.add_row("IPv6 address", f"{report.totalIPv6Address}", f"{report.totalMatchIpv6Address}",
                  f"{0 if report.totalIPv6Address == 0 else round(report.totalMatchIpv6Address * 100.0 / report.totalIPv6Address, 2)}%")
    table.add_row("Email address", f"{report.totalEmailAddress}", f"{report.totalMatchEmailNum}",
                  f"{0 if report.totalEmailNum == 0 else round(report.totalMatchEmailNum * 100.0 / report.totalEmailNum, 2)}%")
    table.add_row("Domain", f"{report.totalDomainNum}", f"{report.totalMatchDomain}",
                  f"{0 if report.totalDomainNum == 0 else round(report.totalMatchDomain * 100.0 / report.totalDomainNum, 2)}%")
    table.add_row("Uri", f"{report.totalHTTPNum}", f"{report.totalMatchUri}",
                  f"{0 if report.totalHTTPNum == 0 else round(report.totalMatchUri * 100.0 / report.totalHTTPNum, 2)}%")
    table.add_row("File hash", f"{report.totalFileNum}", f"{report.totalMatchFileHash}",
                  f"{0 if report.totalFileNum == 0 else round(report.totalMatchFileHash * 100.0 / report.totalFileNum, 2)}%")
    console.print(table)


def main():
    """
    cctxpa is a command lien tool for CCTX to parse pcap file and compare with CCTX's Observables
    """
    if len(sys.argv) == 1:
        sys.argv.append('--help')
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default="127.0.0.1", help="CCTX pcap analyser server addresss")
    parser.add_argument('--port', type=int, default=5000, help="CCTX pcap analyser server port")
    parser.add_argument('--https', action='store_true', help="Use https or http")
    parser.add_argument('--register', action='store_true', help="Just do register account")
    parser.add_argument('--path', type=str, default="", help="CCTX pcap analyser server login path")
    parser.add_argument('--progress', action='store_true',
                        help="Print progress, if open, maybe lead slow extract speed.")
    parser.add_argument('-u', '--username', type=str, required=True, help="Username")
    parser.add_argument('-p', '--password', type=str, required=True, help="Password")
    parser.add_argument('-f', '--pcapfile', type=str, default='test.pcap', help="Pcap file need to parse!")
    parser.add_argument('-o', '--outputfile', type=str, default="report.json", help="A file to store output report")
    parser.add_argument('-ocd', '--outputcsvdirectory', type=str, default='outputcsv',
                        help="A directory to store output csv file")
    args = parser.parse_args()

    cctxpa = CCTXPcapAnalyser(args)
    if args.register:
        # do register
        print(cctxpa.doRegister())
    else:
        begin = datetime.datetime.now().timestamp()
        cctxpa.start()
        with open(f"{args.pcapfile}-{args.outputfile}", 'w') as file:
            file.write(json.dumps(cctxpa.report.toDict(), ensure_ascii=False, cls=MyEncoder))
        printReport(cctxpa.report, args)
        print("Duration: ", datetime.datetime.now().timestamp() - begin)


if __name__ == '__main__':
    main()



import base64
import hashlib
import json
from email.header import decode_header
from email.message import Message
from email.utils import parseaddr


class MailFile:
    def __init__(self, fileName, fileType, fileData):
        self.fileName = fileName
        self.fileType = fileType
        self.fileData = fileData


def guess_charset(msg):
    charset = msg.get_charset()
    if charset is None:
        content_type = msg.get('Content-Type', '').lower()
        pos = content_type.find('charset=')
        if pos >= 0:
            charset = content_type[pos + 8:].strip()
    return charset


def decode_str(s):
    value, charset = decode_header(s)[0]
    if charset:
        value = value.decode(charset)
    return value


class Mail:
    def __init__(self):
        self.plain = ""  # 消息内容
        self.html = ""  # html格式的内容
        self.From = ""  # 发件人地址
        self.To = ""  # 收件人地址
        self.Cc = ""  # 抄送地址
        self.Date = ""  # 日期和时间
        self.Subject = ""  # 主题
        self.MessageID = ""  # 消息ID
        self.files = []
        # self.Received = ""  # 传输路径
        # self.ReturnPath = ""  # 回复地址
        # self.DeliveredTo = ""  # 发送地址
        # self.ReplyTo = ""  # 回复地址
        # self.Bcc = ""  # 暗送地址
        # self.MIMEVersion = ""  # MIME 版本
        """
        内容类型（Content-Type），表现形式为：Content-Type: [type]/[subtype]。
            其中 type 的形式为：
            text：用于标准化地表示的文本信息，文本消息可以是多种字符集和或者多种格式的。
            Image：用于传输静态图片数据。
            Audio：用于传输音频或者音声数据。
            Video：用于传输动态影像数据，可以是与音频编辑在一起的视频数据格式。
            Application：用于传输应用程序数据或者二进制数据。
            Message：用于包装一个 E-mail 消息。
            Multipart：用于连接消息体的多个部分构成一个消息，这些部分可以是不同类型的数据。
            其中 subtype 用于指定 type 的详细形式，常用的 subtype 如下所示：
            text/plain（纯文本）
            text/html（HTML 文档）
            application/xhtml+xml（XHTML 文档）
            image/gif（GIF 图像）
            image/jpeg（JPEG 图像）
            image/png（PNG 图像）
            video/mpeg（MPEG 动画）
            application/octet-stream（任意的二进制数据）
            message/rfc822（RFC 822 形式）
            multipart/alternative（HTML 邮件的 HTML 形式和纯文本形式，相同内容使用不同形式表示。）
        """
        self.ContentType = ""  # 内容的类型
        """
        内容传输编码（Content-Transfer-Encoding），指定内容区域使用的字符编码方式。
        通常为：7bit，8bit，binary，quoted-printable，base64。
        """
        self.ContentTransferEncoding = ""  # 内容的传输编码方式

    def __repr__(self):
        fileHash = b""
        if len(self.files) > 0:
            fileHashes = []
            for item in self.files:
                fileHashes.append(hashlib.md5(item.fileData).hexdigest())
            fileHash = ", ".join(fileHashes)
        if isinstance(self.html, bytes):
            self.html = self.html.decode("utf-8")
        if isinstance(fileHash, bytes):
            fileHash = fileHash.decode("utf-8")
        return json.dumps({
            "Subject": self.Subject,
            "From": self.From,
            "To": self.To,
            "Cc": self.Cc,
            "plain": self.plain,
            "html": self.html,
            "files": fileHash,
            "MessageId": self.MessageID
        }, ensure_ascii=False)
        # return f"Mail(Subject={self.Subject}, From={self.From}, To={self.To}, Cc={self.Cc}, plain={self.plain}, " \
        #        f"html={self.html}, files={fileHash}, MessageId={self.MessageID}"

    def _parse(self, msg: Message, indent):
        """
        https://www.liaoxuefeng.com/wiki/897692888725344/967961517614816
        :param msg:
        :param indent:
        :return:
        """
        if indent == 0:
            for header in ['Subject', 'From', 'To', 'Cc', 'Content-Type', 'Content-Transfer-Encoding', 'Date',
                           'Message-ID']:
                value = msg.get(header, '')
                if value:
                    if header == 'Subject':
                        value = decode_str(value)
                    elif header in ['From', 'To']:
                        hdr, addr = parseaddr(value)
                        value = addr
                    else:
                        hdr, addr = parseaddr(value)
                        name = decode_str(hdr)
                        value = u'%s <%s>' % (name, addr)
                    if header == 'Content-Type':
                        self.ContentType = value
                    elif header == 'Content-Transfer-Encoding':
                        self.ContentTransferEncoding = value
                    elif header == 'Message-ID':
                        self.MessageID = value
                    else:
                        setattr(self, header, value)
        if msg.is_multipart():
            parts = msg.get_payload()
            for part in parts:
                self._parse(part, indent + 1)
        else:
            content_type = msg.get_content_type()
            if content_type == 'text/plain' or content_type == 'text/html':
                content = msg.get_payload(decode=True)
                charset = guess_charset(msg)
                if charset:
                    content = content.decode(charset)
                if content_type == 'text/plain':
                    self.plain = content
                else:
                    self.html = content
            else:
                fileName = msg.get_filename()
                encoding = msg.get('Content-Transfer-Encoding', '')
                data = bytes(msg.get_payload(), 'utf-8')
                if encoding == 'base64':
                    data = base64.b64decode(data)
                self.files.append(MailFile(fileName, content_type, data))

    def parse(self, msg):
        self._parse(msg, 0)
